/*
 * N.I.C.H. Monaic - The 2 channel bbs viewer
 *
 * Copyright (C) 2001 Free Software Foundation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
 * See the AUTHORS file for a list of people on the Monaic Team. See
 * the ChangeLog files for a list of changes. These files are distributed
 * with Monaic at http://monaic.sourceforge.net/
 */

#include "monatext.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gdk/gdkkeysyms.h>

/*
  LINE_HEIGHT(): normal line height
  ALINE_HEIGHT(): LINE_HEIGHT() for Ascii art
*/

#define FONT_HEIGHT(font) (font->ascent + font->descent)
#define LINE_HEIGHT(font) (FONT_HEIGHT(font) + 2)
#define ALINE_HEIGHT(font) FONT_HEIGHT(font)
#define ARTICLE_BORDER(font) LINE_HEIGHT(font)

#define ANUMFORMAT "%d   "

static void mona_text_class_init(MonaTextClass *klass);
static void mona_text_init(MonaText* text);
static void mona_text_destroy(GtkObject* object);
static gint mona_text_expose(GtkWidget* widget, GdkEventExpose *event);
static void mona_text_draw(GtkWidget* widget, GdkRectangle* area);
static void mona_text_realize(GtkWidget* widget);
static gint mona_text_motion_notify(GtkWidget* widget, GdkEventMotion* event);
static gint mona_text_leave_notify(GtkWidget* widget, GdkEventCrossing* event);
static gint mona_text_button_press(GtkWidget* widget, GdkEventButton* event);
static gint mona_text_key_release(GtkWidget* widget, GdkEventKey* event);
static void mona_text_draw_string(GdkWindow* window, GdkFont* font,
  GdkGC* gc, gint x, gint y, const gchar* text);
static void mona_text_draw_text(GdkWindow* window, GdkFont* font,
  GdkGC* gc, gint x, gint y, const gchar* text, gint length);
void mona_text_set_adjustments (MonaText* text, GtkAdjustment* hadj,
  GtkAdjustment* vadj);
static void mona_text_adjustment (GtkAdjustment* adjustment, MonaText* text);

static void mona_text_make_position_cache(MonaText* text);
static void mona_text_real_size_allocate(GtkWidget* widget,
  GtkAllocation* allocation);
static void mona_text_size_request(GtkWidget* widget, GtkRequisition *requisition);

gboolean is_in_rect(GdkRectangle* rect, gint x, gint y);
PosPair* mona_text_find_mouse_cursor_rect(MonaText* text, gint x, gint y);
void mona_text_draw_active_link(MonaText* text, PosPair* pospair);
void mona_text_draw_inactive_link(MonaText* text, PosPair* pospair);
void mona_text_set_active_pospair(MonaText* text, PosPair* pospair);
void parse_links_in_article(PosData* posdata, gchar* article, guint start, GdkFont* font);
static GHashTable* make_url_ht();
gint get_url_length(gchar* text);
gchar* parse_html_in_article(gchar* article);
void clear_focus_area(MonaText* text, gint x, gint y, gint w, gint h);

static GtkWidgetClass *parent_class = NULL;
static const GdkColor fgreen = { 8, 30326, 41377, 33924 };
static const GdkColor blue = { 8, 0, 0, 65535 };
static GHashTable* urlht = NULL;

guint mona_text_get_type(void) {
  static guint mona_text_type = 0;

  if(!mona_text_type) {
    static const GtkTypeInfo mona_text_info = {
      "MonaText",
      sizeof(MonaText),
      sizeof(MonaTextClass),
      (GtkClassInitFunc)mona_text_class_init,
      (GtkObjectInitFunc)mona_text_init,
      NULL,
      NULL,
      (GtkClassInitFunc)NULL
    };
    mona_text_type = gtk_type_unique(gtk_widget_get_type(), &mona_text_info);
  }
  return (mona_text_type);
}

GtkWidget* mona_text_new() {
  MonaText *text;

  text = gtk_type_new(mona_text_get_type());

  return GTK_WIDGET(text);
}

static void mona_text_class_init(MonaTextClass *klass) {
  GtkWidgetClass *widget_class;
  GtkObjectClass *object_class;

  parent_class = gtk_type_class(GTK_TYPE_WIDGET);

  widget_class = (GtkWidgetClass*)klass;
  object_class = (GtkObjectClass*)klass;

  object_class->destroy = mona_text_destroy;

  widget_class->expose_event = mona_text_expose;
  widget_class->realize = mona_text_realize;
  widget_class->draw = mona_text_draw;
  widget_class->motion_notify_event = mona_text_motion_notify;
  widget_class->leave_notify_event = mona_text_leave_notify;
  widget_class->button_press_event = mona_text_button_press;
  widget_class->key_release_event = mona_text_key_release;
  widget_class->size_allocate = mona_text_real_size_allocate;
  //widget_class->size_request = mona_text_size_request;
  widget_class->set_scroll_adjustments_signal =
    gtk_signal_new ("set_scroll_adjustments",
                    GTK_RUN_LAST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (MonaTextClass, set_scroll_adjustments),
                    gtk_marshal_NONE__POINTER_POINTER,
                    GTK_TYPE_NONE, 2, GTK_TYPE_ADJUSTMENT, GTK_TYPE_ADJUSTMENT);

  klass->set_scroll_adjustments = mona_text_set_adjustments;

}

static void mona_text_init(MonaText* text) {

  GTK_WIDGET_SET_FLAGS (text, GTK_CAN_FOCUS);
  GTK_WIDGET_SET_FLAGS (text, GTK_CAN_DEFAULT);

  text->article_slist = NULL;
  text->pos_cache = NULL;
  text->active_pospair = NULL;
  text->offset_y = 0;
  text->total_height = 0; /* Will be set by make_position_cache() */

  text->hadj =
    GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
  text->vadj =
    GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
}

static void mona_text_destroy(GtkObject* object) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(MONA_IS_TEXT(object));

  /* FIXME: add destroy codes */
}

static void mona_text_realize(GtkWidget* widget) {
  GtkStyle* style_copied = NULL;
  MonaText* text;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail(widget != NULL);
  g_return_if_fail(MONA_IS_TEXT(widget));

  text = MONA_TEXT(widget);

  GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);

  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask =  (gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK
                            | GDK_KEY_RELEASE_MASK | GDK_KEY_PRESS_MASK
                            | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
                            | GDK_POINTER_MOTION_MASK | GDK_LEAVE_NOTIFY_MASK);
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (widget->parent->window,
                                   &attributes, attributes_mask);

  //widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);
  //gdk_window_set_background (widget->window, BACKGROUND_COLOR (widget));
  gdk_window_show (widget->window);
  gdk_window_set_user_data (widget->window, widget);

  style_copied = gtk_style_copy(gtk_widget_get_style(widget));
  style_copied->fg[1] = fgreen;
  style_copied->fg[2] = blue;
  gtk_widget_set_style(widget, style_copied);
  gtk_widget_ensure_style(widget);

  /* FIXME: text->pos_cache will leak */
  /* font is changed with realize so it needs recalculate */
  if( text->article_slist != NULL )
    mona_text_make_position_cache(text);
}

static void mona_text_draw(GtkWidget* widget, GdkRectangle* area) {
  gint i;
  gint j;
  MonaText* text;
  gint strx = 0;
  gint stry = 0;
  gchar* anumstr = NULL;

  g_return_if_fail(widget != NULL);
  g_return_if_fail(MONA_IS_TEXT(widget));
  g_return_if_fail(area != NULL);

  text = MONA_TEXT(widget);

  clear_focus_area(MONA_TEXT(widget),
    area->x, area->y, area->width, area->height);

  if( text->article_slist == NULL ) return; /* No article */

  g_return_if_fail(g_slist_length(text->article_slist)
                   == g_slist_length(text->pos_cache));

  if(GTK_WIDGET_DRAWABLE(widget)) {
    /* Skip unnessesary article */
    for( i=1; i<g_slist_length(text->article_slist); i++ ) {
      PosData* posdata = (PosData*)g_slist_nth_data(text->pos_cache, i);
      g_assert(posdata != NULL);
      if( posdata->start > text->offset_y ) {
        if( i != 1 ) i--;
        break;
      }
    }
    for(    ; i<g_slist_length(text->article_slist); i++ ) {
      gchar* cursor = NULL;
      gchar* cursor2 = NULL;
      PosData* posdata = NULL;
      article_dat* adat =
        (article_dat*)g_slist_nth_data(text->article_slist, i);
      g_assert(adat != NULL);

      posdata =
        (PosData*)g_slist_nth_data(text->pos_cache, i);
      g_assert(posdata != NULL);

      /* Draw name line */
      strx = 0;
      stry = posdata->start - text->offset_y;

      if( stry > widget->allocation.height )
        break;

      anumstr = g_strdup_printf(ANUMFORMAT, i);
      mona_text_draw_string(widget->window,
        widget->style->font, widget->style->fg_gc[0], strx,
        stry, anumstr);
      strx += gdk_string_width(widget->style->font, anumstr);
      g_free(anumstr);

      mona_text_draw_string(widget->window,
        widget->style->font, widget->style->fg_gc[1], strx,
        stry, adat->name);

      strx += gdk_string_width(widget->style->font, adat->name);
      strx += gdk_char_width(widget->style->font, ' ')*3;
      mona_text_draw_string(widget->window,
        widget->style->font, widget->style->fg_gc[0], strx,
        stry, adat->date);

      /* Draw article */
      strx = 0;
      stry += LINE_HEIGHT(widget->style->font);
      for( cursor = adat->article; cursor != NULL; ) {
        cursor2 = strstr(cursor, "<br>");
        if( cursor2 == NULL) {
          mona_text_draw_string(widget->window,
            widget->style->font, widget->style->fg_gc[0], strx, stry,
            cursor);
          break;
        } else {
          mona_text_draw_text(widget->window,
            widget->style->font, widget->style->fg_gc[0], strx, stry,
            cursor, cursor2-cursor);
          stry += ALINE_HEIGHT(widget->style->font);
          cursor = cursor2 + strlen("<br>");
        }
      }

    if( posdata->pair_slist != NULL
        && g_slist_length(posdata->pair_slist) > 1 ) {
      for(j=1; j<g_slist_length(posdata->pair_slist); j++ ) {
        PosPair* pospair = g_slist_nth_data(posdata->pair_slist, j);
        g_assert(pospair != NULL);
        mona_text_draw_string(widget->window,
          widget->style->font, widget->style->fg_gc[2],
          pospair->rect->x, pospair->rect->y - text->offset_y,
          pospair->text);
      }
    }

    }
  }
}

static gint
mona_text_expose(GtkWidget* widget, GdkEventExpose *event) {
  mona_text_draw(widget, &event->area);

  return 0;
}

static gint
mona_text_key_release(GtkWidget* widget, GdkEventKey* event) {
  MonaText* text;

  g_return_val_if_fail(widget != NULL, 0);
  g_return_val_if_fail(MONA_IS_TEXT(widget), 0);
  g_return_val_if_fail(event != NULL, 0);

  text = MONA_TEXT(widget);

  switch(event->keyval) {
  case GDK_Page_Down:
    if( text->vadj != NULL) {
      gtk_adjustment_set_value(text->vadj,
        text->vadj->value + text->vadj->page_increment);
    }
    break;
  case GDK_Page_Up:
    if( text->vadj != NULL) {
      gtk_adjustment_set_value(text->vadj,
        text->vadj->value - text->vadj->page_increment);
    }
    break;
  case GDK_Down:
    if( text->vadj != NULL) {
      gtk_adjustment_set_value(text->vadj,
        text->vadj->value + LINE_HEIGHT(widget->style->font));
    }
    break;
  case GDK_Up:
    if( text->vadj != NULL) {
      gtk_adjustment_set_value(text->vadj,
        text->vadj->value - LINE_HEIGHT(widget->style->font));
    }
    break; 
  default:
    break;
  }

  return 0;
}

static gint
mona_text_button_press(GtkWidget* widget, GdkEventButton *event) {
  MonaText* text;
  PosPair* pospair;

  g_return_val_if_fail(widget != NULL, 0);
  g_return_val_if_fail(MONA_IS_TEXT(widget), 0);
  g_return_val_if_fail(event != NULL, 0);

  text = MONA_TEXT(widget);

  if( !GTK_WIDGET_HAS_FOCUS(widget) )
    gtk_widget_grab_focus(widget);

  pospair =
    mona_text_find_mouse_cursor_rect(text,
    (int)event->x, (int)event->y + text->offset_y);

  if( pospair != NULL ) {
    if( pospair->type == MONA_LINK_URL ) {
      gchar* urlcmd;
      gchar* urlstr;

      if( strncmp(pospair->text, "ttp:", strlen("ttp:")) == 0 ) {
        urlstr = g_strconcat("h", pospair->text, NULL);
      } else {
        urlstr = g_strdup(pospair->text);
      }

      urlcmd =
	//g_strdup_printf("mozilla -remote 'openURL(%s)'", urlstr);
	g_strdup_printf("galeon -n '%s' &", urlstr);
      system(urlcmd);
      g_free(urlcmd);
      g_free(urlstr);
    } else if ( pospair->type == MONA_LINK_EMAIL ) {
      g_message("EMAIL: %s", pospair->url);
    } else if ( pospair->type == MONA_LINK_ARTICLE ) {
      g_message("ARTICLE: %s", pospair->url);
    }
  }

  mona_text_set_active_pospair(text, pospair);

  return 0;
}

static gint
mona_text_leave_notify(GtkWidget* widget, GdkEventCrossing *event) {
  MonaText* text;

  g_return_val_if_fail(widget != NULL, 0);
  g_return_val_if_fail(MONA_IS_TEXT(widget), 0);
  g_return_val_if_fail(event != NULL, 0);

  text = MONA_TEXT(widget);

//  if( (int)event->x >= 0 && (int)event->x <= widget->allocation.width
//      && (int)event->y >= 0 && (int)event->y <= widget->allocation.height )
//    return 0;

  mona_text_set_active_pospair(text, NULL);

  return 0;
}

static gint
mona_text_motion_notify(GtkWidget* widget, GdkEventMotion *event) {
  PosPair* pospair = NULL;

  g_return_val_if_fail(widget != NULL, 0);
  g_return_val_if_fail(MONA_IS_TEXT(widget), 0);
  g_return_val_if_fail(event != NULL, 0);

  pospair =
    mona_text_find_mouse_cursor_rect(MONA_TEXT(widget),
    (int)event->x, (int)event->y + MONA_TEXT(widget)->offset_y);

  mona_text_set_active_pospair(MONA_TEXT(widget), pospair);

  return 0;
}

void mona_text_set_article_slist(MonaText* text, GSList* article_slist) {
  g_return_if_fail(text != NULL);
  g_return_if_fail(article_slist != NULL);

  text->article_slist = g_slist_copy(article_slist);
  text->pos_cache = NULL; /* FIXME: free the previous cache */

  mona_text_make_position_cache(text);

  if( GTK_WIDGET_REALIZED(GTK_WIDGET(text)) ) {
    gtk_adjustment_set_value(text->vadj, 0);
    gtk_widget_queue_draw(GTK_WIDGET(text));
  }
}

static gint mona_text_count_newline(gchar* buf) {
  gint i = 1;
  gchar* data = buf;
  gchar* cursor;
  g_return_val_if_fail(data != NULL, 0);

  for( cursor = data ; cursor != NULL ; data = cursor+1) {
    //cursor = (gchar*)strstr(data, "\n");
    cursor = (gchar*)strstr(data, "<br>");
    if (cursor == NULL)
      return i;

    i++;
  }

  return i;
}

static void
mona_text_make_position_cache(MonaText* text) {
  gint i;
  guint height = 0;
  GdkFont* font;

  g_return_if_fail(text != NULL);
  g_return_if_fail(MONA_IS_TEXT(text));
  g_return_if_fail(text->article_slist != NULL);

  if( !GTK_WIDGET_REALIZED(GTK_WIDGET(text)) )
    return; /* Font will change when realized */

  font = GTK_WIDGET(text)->style->font;

  text->pos_cache = g_slist_alloc();
  for(i=1; i<g_slist_length(text->article_slist); i++ ) {
    PosData* posdata = g_new(PosData, 1);
    article_dat* adat =
      (article_dat*)g_slist_nth_data(text->article_slist, i);
    g_assert(adat != NULL);
    posdata->start = height;
    posdata->pair_slist = g_slist_alloc();
    if( adat->email != NULL && *(adat->email) != '\0' ) {
      gchar* anumstr = NULL;
      PosPair* pospair = g_new0(PosPair, 1);
      if( posdata->pair_slist == NULL )
        posdata->pair_slist = g_slist_alloc();
      
      anumstr = g_strdup_printf(ANUMFORMAT, i);
      pospair->rect = g_new0(GdkRectangle, 1);
      pospair->rect->x = gdk_string_width(font, anumstr);
      pospair->rect->y = height;
      pospair->rect->height = FONT_HEIGHT(font);
      pospair->rect->width = gdk_string_width(font, adat->name);
      pospair->text = g_strdup(adat->name);
      pospair->type = MONA_LINK_EMAIL;
      pospair->url = g_strdup(adat->email);
      g_slist_append(posdata->pair_slist, pospair);
      g_free(anumstr);
    }
    g_slist_append(text->pos_cache, posdata);
    height += LINE_HEIGHT(font); /* Name line */
    adat->article = parse_html_in_article(adat->article);
    parse_links_in_article(posdata, adat->article, height, font);
    height += /* Article */
      ALINE_HEIGHT(font) * mona_text_count_newline(adat->article);

    height += ARTICLE_BORDER(font); /* Add article border height */
  }

  text->total_height = height;

  text->vadj->upper = (gfloat)text->total_height;
  text->vadj->step_increment = (gfloat)LINE_HEIGHT(font);
  text->vadj->page_increment = (gfloat)GTK_WIDGET(text)->allocation.height;
  text->vadj->page_size = (gfloat)GTK_WIDGET(text)->allocation.height;
}

gboolean
is_in_rect(GdkRectangle* rect, gint x, gint y) {
  g_return_val_if_fail(rect != NULL, FALSE);

  if( x >= rect->x && x <= rect->x + rect->width
     && y >= rect->y && y <= rect->y + rect->height ) {
    return TRUE;
  }

  return FALSE;
}

void
mona_text_draw_active_link(MonaText* text, PosPair* pospair) {
  GtkWidget* widget = NULL;

  g_return_if_fail(text != NULL);
  g_return_if_fail(MONA_IS_TEXT(text));
  g_return_if_fail(pospair != NULL);

  widget = GTK_WIDGET(text);

  gdk_draw_line(widget->window,
    widget->style->fg_gc[2],
    pospair->rect->x,
    pospair->rect->y + pospair->rect->height-1 - text->offset_y,
    //gdk_string_width(widget->style->font, pospair->text),
    pospair->rect->x + pospair->rect->width,
    pospair->rect->y + pospair->rect->height-1 - text->offset_y);
  mona_text_draw_string(widget->window,
    widget->style->font,
    widget->style->fg_gc[2],
    pospair->rect->x,
    pospair->rect->y - text->offset_y, pospair->text);
}

void
mona_text_draw_inactive_link(MonaText* text, PosPair* pospair) {
  GtkWidget* widget;

  g_return_if_fail(text != NULL);
  g_return_if_fail(MONA_IS_TEXT(text));
  g_return_if_fail(pospair != NULL);

  widget = GTK_WIDGET(text);

  gdk_draw_line(widget->window,
    widget->style->bg_gc[0],
    pospair->rect->x,
    pospair->rect->y + pospair->rect->height-1 - text->offset_y,
    //gdk_string_width(widget->style->font, pospair->text),
    pospair->rect->x + pospair->rect->width,
    pospair->rect->y + pospair->rect->height-1 - text->offset_y);
  mona_text_draw_string(widget->window,
    widget->style->font,
    widget->style->fg_gc[2],
    pospair->rect->x,
    pospair->rect->y - text->offset_y, pospair->text);
}

void
mona_text_set_active_pospair(MonaText* text, PosPair* pospair) {
  g_return_if_fail(text != NULL);
  g_return_if_fail(MONA_IS_TEXT(text));

  if ( pospair == NULL ) {
    if( text->active_pospair != NULL ) {
      mona_text_draw_inactive_link(text, text->active_pospair);
      text->active_pospair = NULL;
    }
    return;
  }

  if( text->active_pospair == pospair )
    return;

  if( text->active_pospair != NULL ) {
      mona_text_draw_inactive_link(text, text->active_pospair);
  }
  mona_text_draw_active_link(text, pospair);
  text->active_pospair = pospair;
}

PosPair*
mona_text_find_mouse_cursor_rect(MonaText* text, gint x, gint y) {
  gint i;
  GtkWidget* widget;
  PosData* posdata = NULL;
  PosData* posdata_prev = NULL;
  PosPair* pospair = NULL;

  g_return_val_if_fail(text != NULL, NULL);
  g_return_val_if_fail(MONA_IS_TEXT(text), NULL);

  widget = GTK_WIDGET(text);

  /* Reject kidding cases */
  if( y < 0 || x < 0 )
    return NULL;

  /* No article */
  if( text->pos_cache == NULL
      || g_slist_length(text->pos_cache) == 1 )
    return NULL;

  for( i=1; i<g_slist_length(text->pos_cache); i++ ) {
    posdata = (PosData*)g_slist_nth_data(text->pos_cache, i);
    g_assert(posdata != NULL);
    if( posdata->start >= y ) {
      i--;
      break;
    }
    posdata_prev = posdata;
  } 
  if( posdata->start < y )
    g_print("Debug: Last article?\n");

  if( posdata_prev != NULL )
    posdata = posdata_prev;

  if( posdata->pair_slist == NULL
      || g_slist_length(posdata->pair_slist) == 1 )
    return NULL; /* No link in this article */

  for( i=1; i<g_slist_length(posdata->pair_slist); i++ ) {
    pospair = (PosPair*)g_slist_nth_data(posdata->pair_slist, i);
    g_assert(pospair != NULL);
    if( is_in_rect(pospair->rect, x, y) == TRUE ) /* Bingo! */
      break;
  } 

  if( i == g_slist_length(posdata->pair_slist) )
    return NULL; /* Cursor is not over the link */

  //g_print("pospair->str: %s\n", pospair->text);

  mona_text_set_active_pospair(text, pospair);

  return pospair;
}

static void mona_text_draw_text(GdkWindow* window, GdkFont* font,
  GdkGC* gc, gint x, gint y, const gchar* text, gint length) {
  gchar* tmpbuf;
  g_return_if_fail(length >= -1 );

  if( length == -1 )
    mona_text_draw_string(window, font, gc, x, y, text);

  tmpbuf = g_strndup(text, length);
  mona_text_draw_string(window, font, gc, x, y, tmpbuf);

  g_free(tmpbuf);
}

static void
mona_text_draw_string(GdkWindow* window, GdkFont* font,
  GdkGC* gc, gint x, gint y, const gchar* text) {
  gdk_draw_string(window, font, gc, x, y + font->ascent, text);
}

void
parse_links_in_article(PosData* posdata, gchar* article, guint start, GdkFont* font) {
  gchar** artv = NULL;
  guint start_height = start;
  gint i;
  g_return_if_fail(posdata != NULL);
  g_return_if_fail(article != NULL);
  g_return_if_fail(font != NULL);

  artv = g_strsplit(article , "<br>", -1);
  if( artv == NULL )
    return; /* We can ignore such case */

  for(i=0; artv[i] != NULL; i++ ) {
    gint urllen;
    gchar* cursor = artv[i];
    for( ;*cursor != '\0'; ) {
      if( strncmp(cursor, "http://", strlen("http://")) == 0
        || strncmp(cursor, "ttp://", strlen("ttp://")) == 0 ) {
        PosPair* pospair = g_new0(PosPair, 1);

        urllen = get_url_length(cursor);

        pospair->text = g_strndup(cursor, urllen);
        pospair->rect = g_new0(GdkRectangle, 1);
        pospair->rect->y = start_height;
        pospair->rect->x = gdk_text_width(font, artv[i], cursor-artv[i]);
        pospair->rect->width = gdk_text_width(font, cursor, urllen);
        pospair->rect->height = FONT_HEIGHT(font);

        if( posdata->pair_slist == NULL )
          posdata->pair_slist = g_slist_alloc();

        g_slist_append(posdata->pair_slist, pospair);

        cursor += urllen;
      } else
        cursor++;
    }

    start_height += ALINE_HEIGHT(font);
  }

  if( artv != NULL )
    g_strfreev(artv);
}

gint get_url_length(gchar* text) {
  gint i = 0;
  g_return_val_if_fail(text, 0);
  g_return_val_if_fail(*text, 0);

  if( urlht == NULL )
    urlht = make_url_ht();

  if( strncmp(text, "http://", strlen("http://")) == 0 ) {
    i = strlen("http://");
  } else if( strncmp(text, "ttp://", strlen("ttp://")) == 0 ) {
    i = strlen("ttp://");
  }

  for( ;i<strlen(text); i++ ) {
    /* Acrobatic cast(w */
    gpointer g = (gpointer)(gint32)text[i];
    if( g_hash_table_lookup(urlht, g) == NULL )
      return i;
  }

  return i;
}

static GHashTable* make_url_ht() {
  /* [a-zA-Z0-9] -_?%&=./?~+ */
  gpointer i;
  GHashTable *ht = g_hash_table_new(NULL, NULL);
  for(i=(gpointer)'a'; i<=(gpointer)'z'; i++ )
    g_hash_table_insert(ht, i, (gpointer)1);

  for(i=(gpointer)'A'; i<=(gpointer)'Z'; i++ )
    g_hash_table_insert(ht, i, (gpointer)1);

  for(i=(gpointer)'1'; i<=(gpointer)'9'; i++ )
    g_hash_table_insert(ht, i, (gpointer)1);

  g_hash_table_insert(ht, (gpointer)'0', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'-', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'_', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'?', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'%', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'&', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'=', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'.', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'/', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'~', (gpointer)1);
  g_hash_table_insert(ht, (gpointer)'+', (gpointer)1);

  return ht;
}

gchar* str_replace(gchar* data, gchar* from, gchar* to) {
  gchar* result = NULL;
  gchar** tmpbuf = NULL;

  g_return_val_if_fail(data != NULL, NULL);
  g_return_val_if_fail(from != NULL, data);
  g_return_val_if_fail(to != NULL, data);

  if( strcmp(from, to) == 0 )
    return data;

  tmpbuf = g_strsplit(data, from, -1);
  result = g_strjoinv(to, tmpbuf);

  if( tmpbuf != NULL )
    g_strfreev(tmpbuf);

  g_free(data); /* Should do this? */

  return result;
}


gchar* parse_html_in_article(gchar* article) {
  gchar* buffer;
  gint i,j;
  gboolean intag = FALSE; 
  g_return_val_if_fail(article != NULL, NULL);

  /* Delete all html tags */
  buffer = g_new0(gchar, strlen(article));
  for(i=0,j=0;i<strlen(article);i++) {
    if( intag == FALSE )
      if( strncmp(article+i, "<a ", strlen("<a ")) == 0
          || strncmp(article+i, "</a>", strlen("</a>")) == 0 ) {
        intag = TRUE;
      } else {
        buffer[j] = article[i];
        j++;
      }
    else if( article[i] == '>' ) {
      intag = FALSE;
    }
  }

  g_free(article); /* Should do here? */
  
  buffer = str_replace(buffer, "&gt;", ">");
  buffer = str_replace(buffer, "&lt;", "<");
  buffer = str_replace(buffer, "&nbsp;", " ");
  buffer = str_replace(buffer, "&quot;", "'");
  buffer = str_replace(buffer, "&amp;", "&"); /* Should be last */

  return buffer;
}

void clear_focus_area(MonaText* text, gint x, gint y, gint w, gint h) {
  GtkWidget* widget;

  g_return_if_fail(text != NULL);

  widget = GTK_WIDGET(text);

  gdk_draw_rectangle (widget->window, widget->style->bg_gc[0], TRUE,
                      x, y, w, h);
}

void
mona_text_set_adjustments (MonaText* text, GtkAdjustment * hadj,
  GtkAdjustment *vadj)
{
  g_return_if_fail(text != NULL);
  g_return_if_fail(MONA_IS_TEXT(text));

  /* FIXME: free the old hadj, vadj */
  if( hadj != NULL ) {
    text->hadj = hadj;
  }
  if( vadj != NULL ) {
    text->vadj = vadj;
    gtk_signal_connect(GTK_OBJECT(text->vadj), "changed",
                       (GtkSignalFunc)mona_text_adjustment, text);
    gtk_signal_connect(GTK_OBJECT(text->vadj), "value_changed",
                       (GtkSignalFunc)mona_text_adjustment, text);
    mona_text_adjustment (vadj, text);
  }
}


static void
mona_text_adjustment (GtkAdjustment* adj, MonaText* text) {
  g_return_if_fail (adj != NULL);
  g_return_if_fail (GTK_IS_ADJUSTMENT (adj));
  g_return_if_fail (text != NULL);
  g_return_if_fail (MONA_IS_TEXT (text));

  if( adj == text->vadj ) {
    text->offset_y = (guint)adj->value;
    if( text->offset_y >
        text->total_height - GTK_WIDGET(text)->allocation.height)
      text->offset_y =
        text->total_height - GTK_WIDGET(text)->allocation.height; 
    gtk_widget_queue_draw(GTK_WIDGET(text));
  }
}

static void
mona_text_real_size_allocate (GtkWidget     *widget,
                               GtkAllocation *allocation)
{
  MonaText *text;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));
  g_return_if_fail (MONA_IS_TEXT(widget));

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget) &&
      !GTK_WIDGET_NO_WINDOW (widget))
     {
        gdk_window_move_resize (widget->window,
                                allocation->x, allocation->y,
                                allocation->width, allocation->height);
     }

  text = MONA_TEXT(widget);
  if( text->vadj != NULL ) {
    text->vadj->upper = (gfloat)text->total_height;
    text->vadj->page_increment = (gfloat)allocation->height;
    text->vadj->page_size = (gfloat)allocation->height;
  }
}

static void mona_text_size_request(GtkWidget* widget,
  GtkRequisition *requisition) {
  g_return_if_fail(widget != NULL);
  g_return_if_fail(MONA_IS_TEXT(widget));
  g_return_if_fail(requisition != NULL);

  requisition->width = 600;
}
